"""Logger for wenxian."""

from __future__ import annotations

import logging


def _get_logger(name):
    """Get a logger."""
    logger = logging.getLogger(name)
    logger.setLevel(logging.WARNING)
    handler = logging.StreamHandler()
    handler.setLevel(logging.WARNING)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


logger = _get_logger(__name__)
"""Global logger for wenxian."""

__all__ = ["logger"]



"""Test cases for the from_identifier function."""

from __future__ import annotations

import pytest

from wenxian.from_identifier import from_identifier

from .cases import TEST_CASES


@pytest.mark.parametrize(
    "identifier, expected",
    [
        *[(test_case.reference.doi, test_case.reference) for test_case in TEST_CASES],
        *[
            # from_identifier accept str
            (str(test_case.pmid), test_case.reference)
            for test_case in TEST_CASES
            if test_case.pmid is not None
        ],
        *[
            (test_case.arxiv, test_case.reference)
            for test_case in TEST_CASES
            if test_case.arxiv is not None
        ],
    ],
)
def test_from_identifier(identifier, expected):
    """Test from_identifier function."""
    assert from_identifier(identifier) == expected



"""Feeder for arXiv."""

from __future__ import annotations

import re
from typing import ClassVar
from xml.etree import ElementTree

from wenxian.feeder.feeder import Feeder
from wenxian.feeder.session import SESSION
from wenxian.reference import Author, Reference


class Arxiv(Feeder):
    """Feeder for arXiv."""

    ARXIV_PATH: ClassVar[dict[str, str]] = {
        "author": r"{http://www.w3.org/2005/Atom}entry/{http://www.w3.org/2005/Atom}author/{http://www.w3.org/2005/Atom}name",
        "title": r"{http://www.w3.org/2005/Atom}entry/{http://www.w3.org/2005/Atom}title",
        "abstract": r"{http://www.w3.org/2005/Atom}entry/{http://www.w3.org/2005/Atom}summary",
        "updated": r"{http://www.w3.org/2005/Atom}entry/{http://www.w3.org/2005/Atom}updated",
    }
    """XPath for arXiv XML."""
    DOI_PREFIX = "10.48550/arXiv."
    """DOI prefix for arXiv."""

    def from_arxiv(self, arxiv: str) -> Reference | None:
        """Fetch a reference from an arXiv identifier."""
        r = SESSION.get("https://export.arxiv.org/api/query", params={"id_list": arxiv})

        tree = ElementTree.fromstring(r.content)

        rets = {}
        for key, path in self.ARXIV_PATH.items():
            if key not in ("author",):
                rr = self._text(tree.find(path))
                if rr is not None:
                    rr = re.sub("[ \n]+", " ", rr)
                rets[key] = rr
        author_tree = tree.findall(self.ARXIV_PATH["author"])
        if author_tree is not None:
            author = []
            for aa in author_tree:
                split_name = self._text(aa).split()
                author.append(
                    Author(first=" ".join(split_name[:-1]), last=split_name[-1])
                )
        else:
            author = None
        if rets["updated"] is not None:
            year = int(rets["updated"].split("-")[0])
        else:
            year = None

        return Reference(
            author=author,
            title=rets["title"].rstrip(".") if rets["title"] is not None else None,
            journal="arXiv",
            year=year,
            annote=rets["abstract"],
            pages=arxiv,
            doi=f"{self.DOI_PREFIX}{arxiv}",
        )

    def from_doi(self, doi: str) -> Reference | None:
        """Fetch a reference from a DOI."""
        if not doi.startswith(self.DOI_PREFIX):
            return None
        return self.from_arxiv(doi[len(self.DOI_PREFIX) :])



"""Feeder for Crossref API."""

from __future__ import annotations

import html

from wenxian.feeder.feeder import Feeder
from wenxian.feeder.session import SESSION
from wenxian.reference import Author, Reference


class Semanticscholar(Feeder):
    """Feeder for Crossref API."""

    def _from_identifier(self, identifier: str) -> Reference | None:
        """Fetch a reference from a identifier."""
        r = SESSION.get(
            f"https://api.semanticscholar.org/graph/v1/paper/{identifier}",
            params={"fields": "title,year,abstract,authors.name,journal,externalIds"},
        )
        if r.status_code == 404:
            # DOI not found
            return None
        res = r.json()
        authors = []
        for author in res["authors"]:
            name = author["name"]
            last = name.split(" ")[-1]
            first = " ".join(name.split(" ")[:-1])
            authors.append(Author(first=first, last=last))
        # journal may be null, e.g.,
        # https://api.semanticscholar.org/graph/v1/paper/10.1103/PhysRevB.109.174106?fields=title,year,abstract,authors.name,journal,externalIds
        if res["journal"] is not None:
            journal = res["journal"]["name"]
            # the journal might be HTML escaped, e.g., Journal of Materials Science &amp; Technology
            # https://api.semanticscholar.org/graph/v1/paper/10.1016/j.jmst.2023.09.059?fields=title,year,abstract,authors.name,journal,externalIds
            journal = html.unescape(journal)
        else:
            journal = None
        return Reference(
            author=authors,
            title=res["title"],
            journal=journal,
            year=res["year"],
            annote=res["abstract"],
            doi=res["externalIds"]["DOI"],
        )

    def from_doi(self, doi: str) -> Reference | None:
        """Fetch a reference from a DOI."""
        return self._from_identifier(doi)

    def from_pmid(self, pmid: str | int) -> Reference | None:
        """Fetch a reference from a PMID."""
        return self._from_identifier(f"PMID:{pmid}")

    def from_arxiv(self, arxiv: str) -> Reference | None:
        """Fetch a reference from an arXiv ID."""
        return self._from_identifier(f"ARXIV:{arxiv}")



"""Tests for the identifier module."""

from __future__ import annotations

import pytest

from wenxian.identifier import Identifier, get_identifier_type


@pytest.mark.parametrize(
    "identifier, expected",
    [
        ("10.1063/5.0155600", Identifier.DOI),
        ("37526163", Identifier.PMID),
        ("2304.09409", Identifier.ARXIV),
    ],
)
def test_get_identifier_type(identifier, expected):
    """Test get_identifier_type()."""
    assert get_identifier_type(identifier) == expected



"""Fetch a reference from an identifier."""

from __future__ import annotations

from wenxian.feeder.arxiv import Arxiv
from wenxian.feeder.crossref import Crossref
from wenxian.feeder.pubmed import Pubmed
from wenxian.feeder.semanticscholar import Semanticscholar
from wenxian.identifier import Identifier, get_identifier_type
from wenxian.reference import Reference


def from_doi(doi: str) -> Reference | None:
    """Fetch a reference from a DOI."""
    # pubmed is the most reliable source
    return (
        Reference()
        | Pubmed().from_doi(doi)
        | Crossref().from_doi(doi)
        | Arxiv().from_doi(doi)
        | Semanticscholar().from_doi(doi)
    )


def from_pmid(pmid: str | int) -> Reference | None:
    """Fetch a reference from a PMID."""
    # no need to fetch from crossref - pubmed usually has all information
    return Reference() | Pubmed().from_pmid(pmid)


def from_arxiv(arxiv: str) -> Reference | None:
    """Fetch a reference from an arXiv identifier."""
    # arxiv api has all information for arxiv papers
    return Reference() | Arxiv().from_arxiv(arxiv)


def from_identifier(identifier: str) -> Reference | None:
    """Fetch a reference from an identifier."""
    identifier_type = get_identifier_type(identifier)
    if identifier_type is None:
        raise ValueError(f"Unknown identifier: {identifier}")
    elif identifier_type == Identifier.DOI:
        return from_doi(identifier)
    elif identifier_type == Identifier.PMID:
        return from_pmid(identifier)
    elif identifier_type == Identifier.ARXIV:
        return from_arxiv(identifier)
    else:
        raise RuntimeError("Unknown identifier type.")



"""Tests for generating BibTeX entries from references."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from .cases import TEST_CASES

if TYPE_CHECKING:
    from wenxian.reference import Reference


@pytest.mark.parametrize(
    "reference, expected", [(cc.reference, cc.expected_bibtex) for cc in TEST_CASES]
)
def test_bibtex(reference: Reference, expected):
    """Test generating BibTeX entries from references."""
    assert reference.bibtex.strip() == expected



"""Feeder base class."""

from __future__ import annotations

from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from xml.etree import ElementTree

    from wenxian.reference import Reference


class Feeder:
    """Feeder base class."""

    def from_doi(self, doi: str) -> Reference | None:
        """Fetch a reference from a DOI."""

    def from_arxiv(self, arxiv: str) -> Reference | None:
        """Fetch a reference from an arXiv identifier."""

    def from_pmid(self, pmid: str | int) -> Reference | None:
        """Fetch a reference from a PubMed identifier."""

    @overload
    def _int(self, string: int) -> int: ...

    @overload
    def _int(self, string: str) -> int | str: ...

    @overload
    def _int(self, string: None) -> None: ...

    def _int(self, string: str | int | None) -> int | str | None:
        """Convert string to int if it is a digit, return None if string is None."""
        if string is None:
            return None
        elif isinstance(string, int):
            return string
        elif isinstance(string, str):
            return int(string) if string.isdigit() else string
        raise ValueError(f"Invalid string: {string}")

    @overload
    def _pages(self, string: str) -> tuple[int]: ...

    @overload
    def _pages(self, string: None) -> None: ...

    def _pages(
        self, string: str | None
    ) -> tuple[int | str] | tuple[int | str, int | str] | None:
        """Convert a page string to a tuple of integers."""
        if string is None:
            return None
        page_list = tuple(self._int(s) for s in string.split("-"))
        if len(page_list) == 1 or len(page_list) == 2:
            return page_list
        raise ValueError(f"Invalid page string: {string}")

    @overload
    def _text(self, node: ElementTree.Element) -> str: ...

    @overload
    def _text(self, node: None) -> None: ...

    def _text(self, node: ElementTree.Element | None) -> str | None:
        """Read text from XML node, return None if node is None."""
        return "".join(node.itertext()).strip() if node is not None else None



"""Test cases."""

from __future__ import annotations

import textwrap
from dataclasses import dataclass

from wenxian.reference import Author, Reference


@dataclass
class ReferenceCase:
    """Test case for a reference."""

    reference: Reference
    expected_bibtex: str
    pmid: int | None = None
    arxiv: str | None = None


TEST_CASES = [
    # mainly from pubmed
    ReferenceCase(
        reference=Reference(
            author=[
                Author(first="Jinzhe", last="Zeng"),
                Author(first="Yujun", last="Tao"),
                Author(first="Timothy J.", last="Giese"),
                Author(first="Darrin M.", last="York"),
            ],
            title="QDπ: A Quantum Deep Potential Interaction Model for Drug Discovery",
            journal="Journal of chemical theory and computation",
            year=2023,
            volume=19,
            issue=4,
            pages=(1261, 1275),
            annote=textwrap.dedent(
                """\
            We report QDπ-v1.0 for modeling the internal energy of drug molecules
            containing H, C, N, and O atoms. The QDπ model is in the form of a
            quantum mechanical/machine learning potential correction (QM/Δ-MLP)
            that uses a fast third-order self-consistent density-functional
            tight-binding (DFTB3/3OB) model that is corrected to a quantitatively
            high-level of accuracy through a deep-learning potential (DeepPot-SE).
            The model has the advantage that it is able to properly treat
            electrostatic interactions and handle changes in charge/protonation
            states. The model is trained against reference data computed at the
            ωB97X/6-31G* level (as in the ANI-1x data set) and compared to several
            other approximate semiempirical and machine learning potentials
            (ANI-1x, ANI-2x, DFTB3, MNDO/d, AM1, PM6, GFN1-xTB, and GFN2-xTB). The
            QDπ model is demonstrated to be accurate for a wide range of intra- and
            intermolecular interactions (despite its intended use as an internal
            energy model) and has shown to perform exceptionally well for relative
            protonation/deprotonation energies and tautomers. An example application
            to model reactions involved in RNA strand cleavage catalyzed by protein
            and nucleic acid enzymes illustrates QDπ has average errors less than
            0.5 kcal/mol, whereas the other models compared have errors over an
            order of magnitude greater. Taken together, this makes QDπ highly
            attractive as a potential force field model for drug discovery.
            """
            )
            .strip()
            .replace("\n", " "),
            doi="10.1021/acs.jctc.2c01172",
        ),
        expected_bibtex=textwrap.dedent(r"""
        @Article{Zeng_JChemTheoryComput_2023_v19_p1261,
            author =   {Jinzhe Zeng and Yujun Tao and Timothy J. Giese and Darrin M. York},
            title =    {{QD{\ensuremath{\pi}}: A Quantum Deep Potential Interaction Model for
                     Drug Discovery}},
            journal =  {J. Chem. Theory Comput.},
            year =     2023,
            volume =   19,
            number =   4,
            pages =    {1261--1275},
            doi =      {10.1021/acs.jctc.2c01172},
            abstract = {We report QD{\ensuremath{\pi}}-v1.0 for modeling the internal energy
                     of drug molecules containing H, C, N, and O atoms. The
                     QD{\ensuremath{\pi}} model is in the form of a quantum
                     mechanical/machine learning potential correction
                     (QM/{\ensuremath{\Delta}}-MLP) that uses a fast third-order self-
                     consistent density-functional tight-binding (DFTB3/3OB) model that is
                     corrected to a quantitatively high-level of accuracy through a deep-
                     learning potential (DeepPot-SE). The model has the advantage that it
                     is able to properly treat electrostatic interactions and handle
                     changes in charge/protonation states. The model is trained against
                     reference data computed at the {\ensuremath{\omega}}B97X/6-31G* level
                     (as in the ANI-1x data set) and compared to several other approximate
                     semiempirical and machine learning potentials (ANI-1x, ANI-2x, DFTB3,
                     MNDO/d, AM1, PM6, GFN1-xTB, and GFN2-xTB). The QD{\ensuremath{\pi}}
                     model is demonstrated to be accurate for a wide range of intra- and
                     intermolecular interactions (despite its intended use as an internal
                     energy model) and has shown to perform exceptionally well for relative
                     protonation/deprotonation energies and tautomers. An example
                     application to model reactions involved in RNA strand cleavage
                     catalyzed by protein and nucleic acid enzymes illustrates
                     QD{\ensuremath{\pi}} has average errors less than 0.5 kcal/mol,
                     whereas the other models compared have errors over an order of
                     magnitude greater. Taken together, this makes QD{\ensuremath{\pi}}
                     highly attractive as a potential force field model for drug discovery.},
        }
    """).strip(),
        pmid=36696673,
    ),
    # no records in PubMed, mainly from crossref
    ReferenceCase(
        reference=Reference(
            author=[
                Author(first="Yuzhi", last="Zhang"),
                Author(first="Haidi", last="Wang"),
                Author(first="Weijie", last="Chen"),
                Author(first="Jinzhe", last="Zeng"),
                Author(first="Linfeng", last="Zhang"),
                Author(first="Han", last="Wang"),
                Author(first="Weinan", last="E"),
            ],
            title="DP-GEN: A concurrent learning platform for the generation of reliable deep learning based potential energy models",
            journal="Computer Physics Communications",
            year=2020,
            volume=253,
            pages=(107206,),
            doi="10.1016/j.cpc.2020.107206",
        ),
        expected_bibtex=textwrap.dedent(r"""
            @Article{Zhang_ComputPhysCommun_2020_v253_p107206,
                author =   {Yuzhi Zhang and Haidi Wang and Weijie Chen and Jinzhe Zeng and Linfeng
                         Zhang and Han Wang and Weinan E},
                title =    {{DP-GEN: A concurrent learning platform for the generation of reliable
                         deep learning based potential energy models}},
                journal =  {Comput. Phys. Commun.},
                year =     2020,
                volume =   253,
                pages =    107206,
                doi =      {10.1016/j.cpc.2020.107206},
            }""").strip(),
    ),
    # arXiv
    ReferenceCase(
        reference=Reference(
            author=[
                Author(first="Yuzhi", last="Zhang"),
                Author(first="Haidi", last="Wang"),
                Author(first="Weijie", last="Chen"),
                Author(first="Jinzhe", last="Zeng"),
                Author(first="Linfeng", last="Zhang"),
                Author(first="Han", last="Wang"),
                Author(first="Weinan", last="E"),
            ],
            title="DP-GEN: A concurrent learning platform for the generation of reliable deep learning based potential energy models",
            journal="arXiv",
            year=2019,
            pages="1910.12690",
            doi="10.48550/arXiv.1910.12690",
            annote=textwrap.dedent("""\
                In recent years, promising deep learning based interatomic potential
                energy surface (PES) models have been proposed that can potentially
                allow us to perform molecular dynamics simulations for large scale
                systems with quantum accuracy. However, making these models truly
                reliable and practically useful is still a very non-trivial task. A
                key component in this task is the generation of datasets used in model
                training. In this paper, we introduce the Deep Potential GENerator
                (DP-GEN), an open-source software platform that implements the
                recently proposed "on-the-fly" learning procedure [Phys. Rev.
                Materials 3, 023804] and is capable of generating uniformly accurate
                deep learning based PES models in a way that minimizes human
                intervention and the computational cost for data generation and model
                training. DP-GEN automatically and iteratively performs three steps:
                exploration, labeling, and training. It supports various popular
                packages for these three steps: LAMMPS for exploration, Quantum
                Espresso, VASP, CP2K, etc. for labeling, and DeePMD-kit for training.
                It also allows automatic job submission and result collection on
                different types of machines, such as high performance clusters and
                cloud machines, and is adaptive to different job management tools,
                including Slurm, PBS, and LSF. As a concrete example, we illustrate
                the details of the process for generating a general-purpose PES model
                for Cu using DP-GEN.
                """)
            .strip()
            .replace("\n", " "),
        ),
        expected_bibtex=textwrap.dedent(r"""
            @Article{Zhang_arXiv_2019_p1910.12690,
                author =   {Yuzhi Zhang and Haidi Wang and Weijie Chen and Jinzhe Zeng and Linfeng
                         Zhang and Han Wang and Weinan E},
                title =    {{DP-GEN: A concurrent learning platform for the generation of reliable
                         deep learning based potential energy models}},
                journal =  {arXiv},
                year =     2019,
                pages =    {1910.12690},
                doi =      {10.48550/arXiv.1910.12690},
                abstract = {In recent years, promising deep learning based interatomic potential
                         energy surface (PES) models have been proposed that can potentially
                         allow us to perform molecular dynamics simulations for large scale
                         systems with quantum accuracy. However, making these models truly
                         reliable and practically useful is still a very non-trivial task. A
                         key component in this task is the generation of datasets used in model
                         training. In this paper, we introduce the Deep Potential GENerator
                         (DP-GEN), an open-source software platform that implements the
                         recently proposed {''}on-the-fly{''} learning procedure [Phys. Rev.
                         Materials 3, 023804] and is capable of generating uniformly accurate
                         deep learning based PES models in a way that minimizes human
                         intervention and the computational cost for data generation and model
                         training. DP-GEN automatically and iteratively performs three steps:
                         exploration, labeling, and training. It supports various popular
                         packages for these three steps: LAMMPS for exploration, Quantum
                         Espresso, VASP, CP2K, etc. for labeling, and DeePMD-kit for training.
                         It also allows automatic job submission and result collection on
                         different types of machines, such as high performance clusters and
                         cloud machines, and is adaptive to different job management tools,
                         including Slurm, PBS, and LSF. As a concrete example, we illustrate
                         the details of the process for generating a general-purpose PES model
                         for Cu using DP-GEN.},
            }
            """).strip(),
        arxiv="1910.12690",
    ),
    # test pii
    ReferenceCase(
        Reference(
            author=[
                Author(first="Jinzhe", last="Zeng"),
                Author(first="Duo", last="Zhang"),
                Author(first="Denghui", last="Lu"),
                Author(first="Pinghui", last="Mo"),
                Author(first="Zeyu", last="Li"),
                Author(first="Yixiao", last="Chen"),
                Author(first="Marián", last="Rynik"),
                Author(first="Li'ang", last="Huang"),
                Author(first="Ziyao", last="Li"),
                Author(first="Shaochen", last="Shi"),
                Author(first="Yingze", last="Wang"),
                Author(first="Haotian", last="Ye"),
                Author(first="Ping", last="Tuo"),
                Author(first="Jiabin", last="Yang"),
                Author(first="Ye", last="Ding"),
                Author(first="Yifan", last="Li"),
                Author(first="Davide", last="Tisi"),
                Author(first="Qiyu", last="Zeng"),
                Author(first="Han", last="Bao"),
                Author(first="Yu", last="Xia"),
                Author(first="Jiameng", last="Huang"),
                Author(first="Koki", last="Muraoka"),
                Author(first="Yibo", last="Wang"),
                Author(first="Junhan", last="Chang"),
                Author(first="Fengbo", last="Yuan"),
                Author(first="Sigbjørn Løland", last="Bore"),
                Author(first="Chun", last="Cai"),
                Author(first="Yinnian", last="Lin"),
                Author(first="Bo", last="Wang"),
                Author(first="Jiayan", last="Xu"),
                Author(first="Jia-Xin", last="Zhu"),
                Author(first="Chenxing", last="Luo"),
                Author(first="Yuzhi", last="Zhang"),
                Author(first="Rhys E. A.", last="Goodall"),
                Author(first="Wenshuo", last="Liang"),
                Author(first="Anurag Kumar", last="Singh"),
                Author(first="Sikai", last="Yao"),
                Author(first="Jingchao", last="Zhang"),
                Author(first="Renata", last="Wentzcovitch"),
                Author(first="Jiequn", last="Han"),
                Author(first="Jie", last="Liu"),
                Author(first="Weile", last="Jia"),
                Author(first="Darrin M.", last="York"),
                Author(first="Weinan", last="E"),
                Author(first="Roberto", last="Car"),
                Author(first="Linfeng", last="Zhang"),
                Author(first="Han", last="Wang"),
            ],
            title="DeePMD-kit v2: A software package for deep potential models",
            journal="The Journal of chemical physics",
            year=2023,
            volume=159,
            issue=5,
            pages="054801",
            doi="10.1063/5.0155600",
            annote=textwrap.dedent(
                """\
                DeePMD-kit is a powerful open-source software package that facilitates
                molecular dynamics simulations using machine learning potentials known
                as Deep Potential (DP) models. This package, which was released in
                2017, has been widely used in the fields of physics, chemistry,
                biology, and material science for studying atomistic systems. The
                current version of DeePMD-kit offers numerous advanced features, such
                as DeepPot-SE, attention-based and hybrid descriptors, the ability to
                fit tensile properties, type embedding, model deviation, DP-range
                correction, DP long range, graphics processing unit support for
                customized operators, model compression, non-von Neumann molecular
                dynamics, and improved usability, including documentation, compiled
                binary packages, graphical user interfaces, and application
                programming interfaces. This article presents an overview of the
                current major version of the DeePMD-kit package, highlighting its
                features and technical details. Additionally, this article presents a
                comprehensive procedure for conducting molecular dynamics as a
                representative application, benchmarks the accuracy and efficiency of
                different models, and discusses ongoing developments.
                """
            )
            .strip()
            .replace("\n", " "),
        ),
        expected_bibtex=textwrap.dedent(r"""
            @Article{Zeng_JChemPhys_2023_v159_p054801,
                author =   {Jinzhe Zeng and Duo Zhang and Denghui Lu and Pinghui Mo and Zeyu Li
                         and Yixiao Chen and Mari{\'a}n Rynik and Li'ang Huang and Ziyao Li and
                         Shaochen Shi and Yingze Wang and Haotian Ye and Ping Tuo and Jiabin
                         Yang and Ye Ding and Yifan Li and Davide Tisi and Qiyu Zeng and Han
                         Bao and Yu Xia and Jiameng Huang and Koki Muraoka and Yibo Wang and
                         Junhan Chang and Fengbo Yuan and Sigbj{\o}rn L{\o}land Bore and Chun
                         Cai and Yinnian Lin and Bo Wang and Jiayan Xu and Jia-Xin Zhu and
                         Chenxing Luo and Yuzhi Zhang and Rhys E. A. Goodall and Wenshuo Liang
                         and Anurag Kumar Singh and Sikai Yao and Jingchao Zhang and Renata
                         Wentzcovitch and Jiequn Han and Jie Liu and Weile Jia and Darrin M.
                         York and Weinan E and Roberto Car and Linfeng Zhang and Han Wang},
                title =    {{DeePMD-kit v2: A software package for deep potential models}},
                journal =  {J. Chem. Phys.},
                year =     2023,
                volume =   159,
                number =   5,
                pages =    054801,
                doi =      {10.1063/5.0155600},
                abstract = {DeePMD-kit is a powerful open-source software package that facilitates
                         molecular dynamics simulations using machine learning potentials known
                         as Deep Potential (DP) models. This package, which was released in
                         2017, has been widely used in the fields of physics, chemistry,
                         biology, and material science for studying atomistic systems. The
                         current version of DeePMD-kit offers numerous advanced features, such
                         as DeepPot-SE, attention-based and hybrid descriptors, the ability to
                         fit tensile properties, type embedding, model deviation, DP-range
                         correction, DP long range, graphics processing unit support for
                         customized operators, model compression, non-von Neumann molecular
                         dynamics, and improved usability, including documentation, compiled
                         binary packages, graphical user interfaces, and application
                         programming interfaces. This article presents an overview of the
                         current major version of the DeePMD-kit package, highlighting its
                         features and technical details. Additionally, this article presents a
                         comprehensive procedure for conducting molecular dynamics as a
                         representative application, benchmarks the accuracy and efficiency of
                         different models, and discusses ongoing developments.},
            }""").strip(),
    ),
    # semanticscholar abstract
    ReferenceCase(
        reference=Reference(
            author=[
                Author(first="Jinzhe", last="Zeng"),
                Author(first="Linfeng", last="Zhang"),
                Author(first="Han", last="Wang"),
                Author(first="Tong", last="Zhu"),
            ],
            title="Exploring the Chemical Space of Linear Alkane Pyrolysis via Deep Potential GENerator",
            journal="Energy Fuels",
            year=2021,
            volume=35,
            issue=1,
            pages=(762, 769),
            annote=textwrap.dedent("""\
                Reactive molecular dynamics (MD) simulation is a powerful tool to study the reaction
                mechanism of complex chemical systems. Central to the method is the potential energy
                surface (PES) that can desc...""")
            .strip()
            .replace("\n", " "),
            doi="10.1021/acs.energyfuels.0c03211",
        ),
        expected_bibtex=textwrap.dedent(r"""
            @Article{Zeng_EnergyFuels_2021_v35_p762,
                author =   {Jinzhe Zeng and Linfeng Zhang and Han Wang and Tong Zhu},
                title =    {{Exploring the Chemical Space of Linear Alkane Pyrolysis via Deep
                         Potential GENerator}},
                journal =  {Energy Fuels},
                year =     2021,
                volume =   35,
                number =   1,
                pages =    {762--769},
                doi =      {10.1021/acs.energyfuels.0c03211},
                abstract = {Reactive molecular dynamics (MD) simulation is a powerful tool to
                         study the reaction mechanism of complex chemical systems. Central to
                         the method is the potential energy surface (PES) that can desc...},
            }""").strip(),
    ),
]



"""Requests session with rate limiting."""

from __future__ import annotations

from pyrate_limiter import Duration, Limiter, RequestRate
from requests import Session
from requests.adapters import HTTPAdapter, Retry
from requests_ratelimiter import LimiterAdapter

SESSION = Session()

# retry logic
retries = Retry(
    total=5,
    backoff_factor=0.1,
    status_forcelist=[
        429,  # Too Many Requests
        500,  # Internal Server Error
        502,  # bad gateway server error
        503,  # the server is currently unable to handle the incoming requests
        504,  # Gateway Timeout
    ],
)

# https://www.ncbi.nlm.nih.gov/books/NBK25497/
adapter_ncbi = LimiterAdapter(per_second=3, max_retries=retries)
SESSION.mount("https://www.ncbi.nlm.nih.gov/pmc/utils/", adapter_ncbi)
SESSION.mount("https://eutils.ncbi.nlm.nih.gov/", adapter_ncbi)
# Crossref rate limit is not documented but discussed in
# https://github.com/CrossRef/rest-api-doc/issues/196
adapter_crossref = LimiterAdapter(per_second=50, max_retries=retries)
SESSION.mount("https://api.crossref.org/", adapter_crossref)
# Arxiv: https://info.arxiv.org/help/api/tou.html
adapter_arxiv = LimiterAdapter(
    limiter=Limiter(RequestRate(1, Duration.SECOND * 3)), burst=1, max_retries=retries
)
SESSION.mount("https://export.arxiv.org/api", adapter_arxiv)
# https://www.semanticscholar.org/product/api
adapter_semanticscholar = LimiterAdapter(per_second=1, max_retries=retries)
SESSION.mount("https://api.semanticscholar.org/", adapter_semanticscholar)

SESSION.mount("https://", HTTPAdapter(max_retries=retries))

__all__ = ["SESSION"]



"""Reference class for generating BibTeX entries."""

from __future__ import annotations

import re
import textwrap
from dataclasses import dataclass

import unidecode
from pyiso4.ltwa import Abbreviate
from pylatexenc.latexencode import unicode_to_latex

abbreviator = Abbreviate.create()


@dataclass
class Author:
    """Author name."""

    first: str | None
    last: str | None
    suffix: str | None = None

    def __str__(self) -> str:
        """Generate a name string."""
        if self.first is None or self.last is None:
            raise ValueError(
                f"First name ({self.first}) or last name ({self.last}) is missing."
            )
        if self.suffix is not None:
            return f"{self.first} {{{self.last} {self.suffix}}}"
        elif " " in self.last:
            return f"{self.first} {{{self.last}}}"
        return f"{self.first} {self.last}"


@dataclass
class Reference:
    """A reference to a scholarly article."""

    author: list[Author] | None = None
    title: str | None = None
    journal: str | None = None
    year: int | None = None
    volume: int | str | None = None
    issue: int | str | None = None
    pages: tuple[int] | tuple[int, int] | str | None = None
    annote: str | None = None
    doi: str | None = None

    @property
    def journal_abbr(self) -> str | None:
        """Abbreviated journal name."""
        if self.journal is None:
            return None
        if self.journal == "arXiv":
            # special case
            return "arXiv"
        abbr = abbreviator(self.journal.title(), remove_part=True)
        # workaround to fix the missing E, e.g. Phys. Rev. E
        # https://github.com/pierre-24/pyiso4/issues/13
        # Example: 10.1103/PhysRevE.108.055310
        if self.journal.title().endswith(" E") and not abbr.endswith(" E"):
            abbr += " E"
        if abbr.replace(",", ".") == self.journal.title():
            # assume it is already abbreviated, cannot handle cases like "J. Chem. Phys."
            # https://github.com/pierre-24/pyiso4/issues/11
            # Example: 10.1021/acs.jpcc.3c05522
            # when it contains ., it may not be abbreviated
            # Example: 10.1021/acs.jpcb.3c05928
            return self.journal.title()
        return abbr

    @property
    def key(self) -> str:
        """Generate a BibTeX key."""
        if self.author is None or len(self.author) == 0:
            raise ValueError("No author is found.")
        if self.author[0].last is None:
            raise ValueError("The first author has no last name.")
        journal_abbr = self.journal_abbr
        if journal_abbr is None:
            raise ValueError("No journal is found.")
        if self.pages is None:
            first_page: str | int | None = None
        elif isinstance(self.pages, tuple):
            first_page = self.pages[0]
        else:
            first_page = self.pages
        key = "{last}_{journal}".format(
            last=unidecode.unidecode(self.author[0].last).replace(" ", ""),
            journal=re.sub(r"[\ \-\.:,]", "", journal_abbr),
        )
        if self.year is not None:
            key += f"_{self.year}"
        if self.volume is not None:
            key += f"_v{self.volume}"
        if first_page is not None:
            key += f"_p{first_page}"
        # special characters are not allowed anywhere in the key, not only in journal
        # https://tex.stackexchange.com/a/96918/262109
        key = re.sub(r"""[\ "#%'\(\),\=\{\}]""", "", key)
        return key

    @property
    def bibtex(self) -> str:
        """Generate a BibTeX entry."""
        if self.author is None:
            author_string = None
        else:
            author_string = " and ".join(str(aa) for aa in self.author)
        if self.pages is None:
            page_string = None
        elif isinstance(self.pages, tuple):
            page_string = "--".join(str(x) for x in self.pages)
        else:
            page_string = str(self.pages)

        # See https://us.mirrors.cicku.me/ctan/macros/latex/contrib/biblatex/doc/biblatex.pdf
        data = {
            "author": author_string,
            "title": self.title,
            "journal": self.journal_abbr,
            "year": self.year,
            "volume": self.volume,
            "number": self.issue,
            "pages": page_string,
            "doi": self.doi,
            "abstract": self.annote,
        }
        start = f"@Article{{{self.key},"
        end = "}\n"
        items = [start]
        for key, value in data.items():
            if value is None:
                continue
            if isinstance(value, int) or (isinstance(value, str) and value.isdigit()):
                valuestr = str(value)
            else:
                valuestr = ("\n" + " " * 13).join(
                    textwrap.wrap(
                        unidecode.unidecode(
                            unicode_to_latex(
                                value,
                                non_ascii_only=False,
                                replacement_latex_protection="braces-all",
                            )
                        ),
                        70,
                    )
                )
                if key == "title":
                    valuestr = f"{{{{{valuestr}}}}}"
                else:
                    valuestr = f"{{{valuestr}}}"
            valuestr += ","
            keystr = " " * 4 + (key + " =").ljust(11, " ")
            items.append(keystr + valuestr)
        items.append(end)
        return "\n".join(items)

    def __or__(self, other: Reference | None) -> Reference:
        """Combine two references."""
        if other is None:
            return self
        return Reference(
            author=self.author or other.author,
            title=self.title or other.title,
            journal=self.journal or other.journal,
            year=self.year or other.year,
            volume=self.volume or other.volume,
            issue=self.issue or other.issue,
            pages=self.pages or other.pages,
            annote=self.annote or other.annote,
            doi=self.doi or other.doi,
        )

    def is_empty(self) -> bool:
        """Check if the reference is empty."""
        return all(
            (
                self.author is None,
                self.title is None,
                self.journal is None,
                self.year is None,
                self.volume is None,
                self.issue is None,
                self.pages is None,
                self.annote is None,
                self.doi is None,
            )
        )



"""Feeder for PubMed."""

from __future__ import annotations

from typing import ClassVar
from xml.etree import ElementTree

from wenxian import __email__, __tool__
from wenxian.feeder.feeder import Feeder
from wenxian.feeder.session import SESSION
from wenxian.reference import Author, Reference


class Pubmed(Feeder):
    """Feeder for PubMed."""

    def _doi2pmid_pmc(self, doi: str) -> str | None:
        """Convert DOI to PMID using PMC database.

        Parameters
        ----------
        doi : str
            A DOI string.

        Returns
        -------
        str | None
            A PMID string if found or None if not found.
        """
        r = SESSION.get(
            "https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/",
            params={"tool": __tool__, "email": __email__, "ids": doi, "format": "json"},
        )
        res = r.json()
        if res["status"] == "error":
            return None
        records = res["records"]

        if records and "pmid" in records[0]:
            return records[0]["pmid"]
        return None

    def _doi2pmid_search(self, doi: str) -> str | None:
        """Convert DOI to PMID using PubMed search.

        The returned PMID may be incorrect if the reference is not in the
        database. Need to validate it.

        Parameters
        ----------
        doi : str
            A DOI string.

        Returns
        -------
        str | None
            A PMID string if found or None if not found.
        """
        r = SESSION.get(
            "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi",
            params={
                "tool": __tool__,
                "email": __email__,
                "db": "pubmed",
                "term": doi,
                "retmode": "json",
                "retmax": "1",
            },
        )
        records = r.json()["esearchresult"]["idlist"]

        if records:
            return records[0]
        return None

    PUBMED_PATH: ClassVar[dict[str, str]] = {
        "author": "PubmedArticle/MedlineCitation/Article/AuthorList/Author",
        "title": "PubmedArticle/MedlineCitation/Article/ArticleTitle",
        "abstract": "PubmedArticle/MedlineCitation/Article/Abstract/AbstractText",
        "journal": "PubmedArticle/MedlineCitation/Article/Journal/Title",
        "volume": "PubmedArticle/MedlineCitation/Article/Journal/JournalIssue/Volume",
        "issue": "PubmedArticle/MedlineCitation/Article/Journal/JournalIssue/Issue",
        "year": "PubmedArticle/MedlineCitation/Article/Journal/JournalIssue/PubDate/Year",
        "pages": "PubmedArticle/MedlineCitation/Article/Pagination/MedlinePgn",
        "doi": "PubmedArticle/PubmedData/ArticleIdList/ArticleId[@IdType='doi']",
        "pii": "PubmedArticle/MedlineCitation/Article/ELocationID[@EIdType='pii']",
    }
    """XPath for PubMed XML."""

    def from_doi(self, doi: str) -> Reference | None:
        """Fetch a reference from a DOI."""
        # get PMID. PMC is more reliable than search
        pmid = self._doi2pmid_pmc(doi)
        if pmid is None:
            pmid = self._doi2pmid_search(doi)
        if pmid is None:
            # not found
            return None
        return self._from_pmid(pmid, validate_doi=doi)

    def from_pmid(self, pmid: str | int) -> Reference | None:
        """Fetch a reference from a PMID."""
        return self._from_pmid(pmid)

    def _from_pmid(
        self, pmid: str | int, validate_doi: str | None = None
    ) -> Reference | None:
        r = SESSION.get(
            "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi",
            params={
                "tool": __tool__,
                "email": __email__,
                "db": "pubmed",
                "id": str(pmid),
                "format": "xml",
            },
        )
        tree = ElementTree.fromstring(r.content)
        fetched_doi = self._text(tree.find(self.PUBMED_PATH["doi"]))
        if validate_doi is not None and fetched_doi != validate_doi:
            # DOI not match, ignore
            return None
        rets = {}
        for key, path in self.PUBMED_PATH.items():
            if key not in ("author",):
                rets[key] = self._text(tree.find(path))

        # specially handle Physical chemistry chemical physics : PCCP
        # DOI: 10.1039/d4cp00997e
        if rets["journal"] == "Physical chemistry chemical physics : PCCP":
            rets["journal"] = "Physical chemistry chemical physics"

        author_tree = tree.findall(self.PUBMED_PATH["author"])
        if author_tree is not None:
            author = []
            for aa in author_tree:
                first = self._text(aa.find("ForeName"))
                if first is not None:
                    # add period if it is a single letter, e.g., Darrin M -> Darrin M.
                    first = " ".join(
                        f"{x}." if len(x) == 1 else x for x in first.split()
                    )
                author.append(
                    Author(
                        first=first,
                        last=self._text(aa.find("LastName")),
                        suffix=self._text(aa.find("Suffix")),
                    )
                )
        else:
            author = None
        year = self._int(rets["year"])
        if year is not None:
            assert isinstance(year, int)
        return Reference(
            author=author,
            title=rets["title"].rstrip(".") if rets["title"] is not None else None,
            journal=rets["journal"],
            year=year,
            volume=self._int(rets["volume"]),
            issue=self._int(rets["issue"]),
            pages=self._pages(rets["pages"]) or rets["pii"],
            annote=rets["abstract"],
            doi=fetched_doi,
        )



"""Test command line interface."""

from __future__ import annotations

import subprocess
import sys
import tempfile

from .cases import TEST_CASES


def test_cli_from_to_stdout():
    """Test wenxian from DOI to stdout."""
    case = TEST_CASES[0]
    out = subprocess.check_output(
        [sys.executable, "-m", "wenxian", "from", case.reference.doi], text=True
    )
    assert out.strip() == case.expected_bibtex.strip()


def test_cli_from_to_file():
    """Test wenxian from DOI to a certain file."""
    case = TEST_CASES[0]
    with tempfile.NamedTemporaryFile("w+") as f:
        subprocess.check_call(
            [sys.executable, "-m", "wenxian", "from", case.reference.doi, "-o", f.name],
            text=True,
        )
        f.seek(0)
        assert f.read().strip() == case.expected_bibtex.strip()


def test_cli_ignore_errors():
    """Test wenxian from DOI to stdout."""
    # expected success
    subprocess.check_call(
        [
            sys.executable,
            "-m",
            "wenxian",
            "from",
            "this_is_not_an_identifier",
            "--ignore-errors",
        ],
        text=True,
    )
    # expected failure
    retcode = subprocess.call(
        [sys.executable, "-m", "wenxian", "from", "this_is_not_an_identifier"],
        text=True,
    )
    assert retcode == 1



"""Identifier type and regex."""

from __future__ import annotations

import re
from enum import Enum


class Identifier(Enum):
    """Identifier type."""

    DOI = "DOI"
    PMID = "PMID"
    ARXIV = "arXiv"


REGEX: dict[Identifier, re.Pattern] = {
    Identifier.DOI: re.compile(r"10\.\d{4,9}/[-._;()/:A-Z0-9]+", re.IGNORECASE),
    Identifier.PMID: re.compile(r"\b\d{8}\b"),
    Identifier.ARXIV: re.compile(r"\d{4}\.\d{4,5}(v\d+)?"),
}
"""Regex for identifiers."""


def get_identifier_type(identifier: str) -> Identifier | None:
    """Get the type of an identifier."""
    for id_type, regex in REGEX.items():
        if regex.match(identifier):
            return id_type
    return None



"""Nox configuration file."""

from __future__ import annotations

import nox


@nox.session
def tests(session: nox.Session) -> None:
    """Run test suite with pytest."""
    session.install("-e.[test]")
    session.run(
        "pytest",
        "--cov",
        "--cov-config",
        "pyproject.toml",
        "--cov-report",
        "term",
        "--cov-report",
        "xml",
    )



"""Main entry point for the command line interface."""

from __future__ import annotations

import argparse
import sys

from wenxian.from_identifier import from_identifier
from wenxian.logger import logger


def cmd_from(
    *,
    IDENTIFIER: list[str],
    output: str | None = None,
    ignore_errors: bool = False,
    **kwargs,
):
    """Generate BibTeX from a identifier."""
    buff = []
    references = []
    for identifier in IDENTIFIER:
        try:
            ref = from_identifier(identifier.strip())
        except Exception as e:
            msg = f"Failed to fetch reference from {identifier}: {e}"
            if ignore_errors:
                logger.exception(msg)
                continue
            else:
                raise ValueError(msg) from e
        if ref is None or ref.is_empty():
            msg = f"Failed to fetch reference from {identifier}"
            if ignore_errors:
                logger.error(msg)
                continue
            else:
                raise ValueError(msg)
        references.append(ref)
        buff.append(ref.bibtex)
    if output is None:
        sys.stdout.write("\n".join(buff))
        return
    if output == 0:
        if len(references) == 1:
            output = f"{references[0].key}.bib"
        else:
            output = "references.bib"
    with open(output, "w") as f:
        f.write("\n".join(buff))


def main_parser() -> argparse.ArgumentParser:
    """Create the main argument parser."""
    parser = argparse.ArgumentParser(description="Generate BibTeX.")
    subparsers = parser.add_subparsers(dest="command", required=True)
    parser_from = subparsers.add_parser(
        "from", help="Generate BibTeX from a identifier."
    )
    parser_from.add_argument(
        "IDENTIFIER",
        type=str,
        nargs="+",
        help="Identifier. Support DOI, PMID, and arXiv ID.",
    )
    parser_from.add_argument(
        "-o",
        "--output",
        type=str,
        nargs="?",
        const=0,
        default=None,
        help=(
            "Output file. If not specified, print to stdout. If specified without a value, print"
            " to a file with item key (for a single entry) or references.bib (for multiple entries)."
        ),
    )
    parser_from.add_argument(
        "--ignore-errors",
        action="store_true",
        help="Ignore errors and continue processing the rest identifiers.",
    )
    parser_from.set_defaults(func=cmd_from)
    return parser


def main():
    """Execute main entry point."""
    parser = main_parser()
    args = parser.parse_args()
    args.func(**vars(args))


if __name__ == "__main__":
    main()



"""Feeder for Crossref API."""

from __future__ import annotations

import html

from wenxian.feeder.feeder import Feeder
from wenxian.feeder.session import SESSION
from wenxian.reference import Author, Reference


class Crossref(Feeder):
    """Feeder for Crossref API."""

    def from_doi(self, doi: str) -> Reference | None:
        """Fetch a reference from a DOI."""
        r = SESSION.get(
            f"https://api.crossref.org/works/{doi}",
        )
        if r.status_code == 404:
            # DOI not found
            return None
        res = r.json()

        m = res["message"]
        # title
        if "title" in m:
            title = m["title"][0]
        else:
            title = None
        # author
        if "author" in m:
            author = []
            for aa in m["author"]:
                author.append(Author(first=aa["given"], last=aa["family"]))
        else:
            author = None
        # volume & issue
        volume = m.get("volume")
        issue = m.get("issue")
        # page
        if "page" in m:
            page = m["page"]
        elif "article-number" in m:
            page = m["article-number"]
        else:
            page = None
        # abstract
        abstract = m.get("abstract")

        # year
        if "published-print" in m:
            year = m["published-print"]["date-parts"][0][0]
        elif "published-online" in m:
            year = m["published-online"]["date-parts"][0][0]
        else:
            year = None
        # journal
        if m.get("short-container-title"):
            journal = m["short-container-title"][0]
            # while not documented, the journal might be HTML escaped, e.g., Journal of Materials Science &amp; Technology
            # https://api.crossref.org/works/10.1016/j.jmst.2023.09.059
            journal = html.unescape(journal)
        else:
            journal = None
        return Reference(
            author=author,
            title=title,
            journal=journal,
            year=self._int(year),
            volume=self._int(volume),
            issue=self._int(issue),
            pages=self._pages(page),
            annote=abstract,
            doi=doi,
        )
